/**
 * Angular directive to convert JSON into HTML tree. Inspired by Brian Park's 
 * MIT Licensed "angular-json-human.js" which turns JSON to HTML tables.
 * 
 *  Extended for trees by Eben Haber at Couchbase.
 *  
 *  This class takes a JS object or JSON string, and displays it as an HTML
 *  list, which object members indented. This is similar to pretty-printing 
 *  JSON, but is more compact (no braces or commas), and permits using colors
 *  to highlight field names vs. values, and one line from the next.
 *  
 *  For usage, see the header to mn-json-table.
 */
/* global _, angular */
'use strict';
var treeModule = angular.module('mnJsonTreeOrig', ['mnJsonTreeOrig.tmpls']);

//
// this factory class is used to recursively expand child objects using the jsonTreeHelper
// directive (see below)
//

treeModule.factory('TreeOrigRecursionHelper', [
  '$compile',
  function ($compile) {
    var RecursionHelper = {
        compile: function (element) {
          var contents = element.contents().remove();
          var compiledContents;
          return function (scope, element) {
            if (!compiledContents) {
              compiledContents = $compile(contents);
            }
            compiledContents(scope, function (clone) {
                element.append(clone);
            });
            var json = scope.json;
            scope.isBoolean = _.isBoolean(json);
            scope.isNumber = _.isNumber(json);
            scope.isString = _.isString(json);
            scope.isPrimitive = scope.isBoolean || scope.isNumber || scope.isString;
            scope.isObject = _.isPlainObject(json);
            scope.isArray = _.isArray(json);
            scope.isEmpty = _.isEmpty(json);
            scope.isPrimFn = function(obj)   {return _.isBoolean(obj) || _.isString(obj) || _.isNumber(obj);};
            scope.numKeys = function(obj)    {return _.keys(obj).length;};
            scope.getKeys = function(obj) {return(Object.keys(obj));};

          };
        }
      };
    return RecursionHelper;
  }
]);

//
// this directive makes sure that we have json data to work with, then puts it
// into the json-tree-root template, which encloses, inside a table, the jsonTreeHelper 
// directive applied to the data. <table json-tree-helper=json></table>
//
// json-tree-helper, below, does recursive magic!
//

treeModule.directive('mnJsonTreeOrig', function () {
  return {
    restrict: 'A',
    scope: { data: '=mnJsonTreeOrig' },
    templateUrl: 'template/mn-json-tree-orig-root.tmpl',
    link: function (scope, linkElement) {
      scope.$watch('data', function (json) {
        if (typeof json === 'string') {
          try {
                scope.json_length = json.length;
            json = JSON.parse(json);
          } catch (e) {
          }
        }
        
        // if the object has an array length one, peek inside
        while (Array.isArray(json) && json.length == 1) {
                json = json[0];
        }
                
        scope.json_length = _.keys(json).length;
        scope.json = json;
        scope.isObject = _.isPlainObject(json);
        scope.isArray = _.isArray(json);
        scope.getKeys = function(obj) {return(Object.keys(obj));};
        
      });
    }
  };
});

//
// this helper directive uses the template below to work recursively:
//  - if it has a primitive, it just outputs the value
//  - if it is an empty object, string, or list, output "(Empty *)"
//  - if it is an array or object, output a table where for each element
//    we show the name and value, with the value tagged with jsonTreeHelper recursively
// gets inserted whenever we find another list or object,
// which gets compiled recursively into more HTML
//

treeModule.directive('mnJsonTreeOrigHelper', [
  'TreeOrigRecursionHelper',
  function (RecursionHelper) {
    return {
      restrict: 'A',
      scope: { json: '=mnJsonTreeOrigHelper' },
      templateUrl: 'template/mn-json-tree-orig.tmpl',
      compile: function (tElem) {
          //console.log("about to compile: " + tElem);
        return RecursionHelper.compile(tElem);
      }
    };
  }
]);

//
// this module exists only to put the templates we use into the template cache
//

angular.module('mnJsonTreeOrig.tmpls', []).run([
  '$templateCache',
  function ($templateCache) {
    'use strict';
    $templateCache
    .put('template/mn-json-tree-orig-root.tmpl', 
         '<div class=ajtd-root ng-class="{ \'ajtd-type-array\': isArray, \'ajtd-type-object\': isObject }" mn-json-tree-orig-helper=json></div>');
    $templateCache
    .put('template/mn-json-tree-orig.tmpl', 
//              '<h4>{{"isBoolean: "+isBoolean + ", isNumber: "+isNumber + ", isString: "+isString + ", isPrimitive: "+ isPrimitive + '+
//              '", isObject: "+isObject + ", isArray: "+ isArray+ ", isEmpty: "+isEmpty + ", json.len: " + numKeys(json)}}</h4>' + 
//      '<br><h4>{{json}}</h4>' +
//      '<br><h4>{{"length: " + json.length}}</h4>' +
//      '<br><h4>{{"isEmpty: "+isEmpty}}</h4>' +
 
                // if nothing is set yet, we need to compile the JSON
                '<div ng-if="isEmpty && numKeys(json) > 0" class=ajtd-object-value mn-json-tree-orig-helper=json></div>' +

                // if it is primitive, output just the value
                '<span ng-if=isPrimitive ng-class="{ \'ajtd-type-bool\': isBoolean, \'ajtd-type-number\': isNumber, \'ajtd-type-string\': isString, \'ajtd-type-array\': isArray, \'ajtd-type-object\': isObject }">{{ json }} <span ng-if="isEmpty && isString" class=ajtd-empty>(Empty String)</span></span>' + 

                // for empty lists and objects, say so
                '<span ng-if="isEmpty && isArray" class=ajtd-empty>(Empty List)</span>' + 
                '<span ng-if="isEmpty && isObject" class=ajtd-empty>(Empty Object)</span>' +

                // for an object or array, ng-repeat over key-values
                '<ul ng-if="!isEmpty && !isPrimitive" ng-class="{ \'ajtd-type-array\': isArray, \'ajtd-type-object\': isObject }">' + 
                  '<li ng-repeat="key in getKeys(json) track by $index">' + 
                    '<div ng-if="isPrimFn(json[key])"><span ng-if="!isArray" class=ajtd-key>{{ key + ": "}}</span><span class=ajtd-value>{{json[key]}}</span></div>' + 
                    '<div ng-if="!isPrimFn(json[key])"><div ng-if="!isArray" class=ajtd-array-key>{{ key }}</div><div class=ajtd-object-value mn-json-tree-orig-helper=json[key]></div></div>' + 
                '</li></ul>');
  }
]);