/**
 * Angular directive to convert JSON into HTML tree. Inspired by Brian Park's 
 * MIT Licensed "angular-json-human.js" which turns JSON to HTML tables.
 * 
 *  Extended for trees by Eben Haber at Couchbase.
 *  
 *  This class takes a JS object or JSON string, and displays it as an HTML
 *  list, which object members indented. This is similar to pretty-printing 
 *  JSON, but is more compact (no braces or commas), and permits using colors
 *  to highlight field names vs. values, and one line from the next.
 *  
 *  For usage, see the header to mn-json-table.
 */
/* global _, angular */
'use strict';
var treeModule = angular.module('mnJsonTreeTweaked', ['mnJsonTreeTweaked.tmpls']);


//this factory class is used to recursively expand child objects using the jsonTreeHelper
//directive (see below)


schemaDisplayModule.factory('TreeTweakedRecursionHelper', ['$compile', function($compile){
  return {
    /**
     * Manually compiles the element, fixing the recursion loop.
     * @param element
     * @param [link] A post-link function, or an object with function(s) registered via pre and post properties.
     * @returns An object containing the linking functions.
     */
    compile: function(element, link){
      // Normalize the link parameter
      if(angular.isFunction(link)){
        link = { post: link };
      }

      // Break the recursion loop by removing the contents
      var contents = element.contents().remove();
      var compiledContents;
      return {
        pre: null,
        /**
         * Compiles and re-adds the contents
         */
        post: function(scope, element){
              //console.log("Inside compile post");
              // Compile the contents
              if(!compiledContents){
                compiledContents = $compile(contents);
              }
              // Re-add the compiled contents to the element
              compiledContents(scope, function(clone){
                element.append(clone);
              });

              // Call the post-linking function, if any
              if(link && link.post){
                link.post.apply(null, arguments);
              }

              scope._ = _;
              scope.j_type = getTypeNameForVal(scope.json);
              scope.removeHashKey = function(name) {
                return(name != '$$hashKey');
              }
              scope.isPrimFn = function(obj)   {return _.isBoolean(obj) || _.isString(obj) || _.isNumber(obj) || !obj;};
              scope.isObjectFn = function(obj) {return _.isPlainObject(obj);}
              scope.isArrayFn = function(obj)  {return _.isArray(obj);}
              scope.getKeys = function(obj)    {return(Object.keys(obj));};
              //console.log("got json: " + json + ", len: " + json.length + ", isEmpty: " + scope.isEmpty);
            }
      };
    }
  };
}]);

var getTypeNameForVal = function(value) {
  if (_.isString(value))
    return('S');
  else if (_.isNumber(value))
    return('N');        
  else if (_.isBoolean(value))
    return('B');
  else if (_.isArray(value))
    return('A');
  else if (_.isPlainObject(value) && _.keys(value).length > 0)
    return('O');
  else if (_.isEpmty(value))
    return('E');
  else
    return('U');
};


//this helper directive uses the template below to work recursively:
//- if it has a primitive, it just outputs the value
//- if it is an empty object, string, or list, output "(Empty *)"
//- if it is an array or object, output a table where for each element
//we show the name and value, with the value tagged with jsonTreeHelper recursively
//gets inserted whenever we find another list or object,
//which gets compiled recursively into more HTML


treeModule.directive('mnJsonTreeTweaked', 
    [
     'TreeTweakedRecursionHelper',
     function (RecursionHelper) {
       return {
         restrict: 'A',
         scope: { json: '=mnJsonTreeTweaked' },
         templateUrl: 'template/mn-json-tree-tweaked.tmpl',
         compile: function (tElem) {
           //console.log("Got tElement: " + JSON.stringify(tElem));

           //console.log("about to compile: " + tElem);
           return RecursionHelper.compile(tElem);
         }
       };
     }
     ]
);


//this module exists only to put the templates we use into the template cache


angular.module('mnJsonTreeTweaked.tmpls', []).
run([
     '$templateCache',
     function ($templateCache) {
       'use strict';
//       $templateCache
//       .put('template/mn-json-tree-root.tmpl', 
//       '<div class=ajtd-root ng-class="{ \'ajtd-type-array\': isArray, \'ajtd-type-object\': isObject }" mn-json-tree-helper=json></div>');
       $templateCache
       .put('template/mn-json-tree-tweaked.tmpl', 
           //'<h3>Staring new iteration isEmpty {{isEmpty}} type: {{j_type}}</h3>' +
           //'<h4>{{"isBoolean: "+isBoolean + ", isNumber: "+isNumber + ", isString: "+isString + ", isPrimitive: "+ isPrimitive + '+
           //'", isObject: "+isObject + ", isArray: "+ isArray+ ", isEmpty: "+isEmpty + ", json.len: " + numKeys(json)}}</h4>' + 
           //'<br><h4>{{json}}</h4>' +
           //'<br><h4>{{"length: " + json.length}}</h4>' +
           //'<br><h4>{{"isArray: "+isArray}}</h4>' +

           // if nothing is set yet, we need to recursively compile the JSON
           //'<div ng-if="json && isEmpty && numKeys(json) > 0" class=ajtd-object-value ng-repeat="f in json" mn-json-tree=f></div>' +

           // if it is primitive, output just the value
           '<div ng-if="j_type != \'A\' && j_type != \'O\'" class="ajtd-type-value">{{ json }}</div>' + 

           // for an array, start a new list and iterate over the elements. If the elements
           // are primitive, output them. Recursion is really expensive, so if we see
           // an object inside the array, output its fields. We only make a recursive call
           // for objects with array/object fields, or for arrays inside arrays.
           
           '<span ng-if="_.isEmpty(json) && j_type == \'A\'" class=ajtd-empty>[]</span>' + 
           '<ul ng-if="j_type == \'A\'" class="ajtd-type-array">' +
           '  <li ng-repeat="obj in json">' +
           '    <div ng-if="isPrimFn(obj)" class=ajtd-value>{{obj}}</div>' + 
           
           // object, iterate through fields
           '    <ul ng-if="_.isPlainObject(obj)" class="ajtd-type-object">' + 
           '      <li ng-repeat="key in getKeys(obj) | filter:removeHashKey">' + 
           '        <div ng-if="isPrimFn(obj[key])"><span class=ajtd-key>{{ key + ": "}}</span><span class=ajtd-value>{{obj[key]}}</span></div>' + 
           '        <div ng-if="_.isArray(obj[key])" class=ajtd-object-value mn-json-tree-tweaked=obj></div>' + 
           '        <div ng-if="_.isPlainObject(obj[key])"><div class=ajtd-array-key>{{ key }}</div>'+
           '          <ul class="ajtd-type-object"><li ng-repeat="subkey in getKeys(obj[key]) | filter:removeHashKey">' + 
           '            <div ng-if="isPrimFn(obj[key][subkey])"><span class=ajtd-key>{{ subkey + ": "}}</span><span class=ajtd-value>{{obj[key][subkey]}}</span></div>' + 
           '            <div ng-if="!isPrimFn(obj[key][subkey])">' +
           '              <div class=ajtd-array-key>{{ key }}</div>' +
           '              <div class=ajtd-object-value mn-json-tree-tweaked=obj[key][subkey]></div>' + 
           '          </li></ul>' +
           '        </div>' + 
           '      </li>' +
           '    </ul>' +
           // array, make recursive call
           '    <div ng-if="_.isArray(obj)" class=ajtd-object-value mn-json-tree-tweaked=obj></div>' + 
           '  </li>' +
           '</ul>' +

           // for an object, ng-repeat over key-values, outputting values by type
           '<span ng-if="_.isEmpty(json) && j_type == \'O\'" class=ajtd-empty>{}</span>' +
           '<ul ng-if="!_.isEmpty(json) && j_type == \'O\'" class="ajtd-type-object">' + 
           '  <li ng-repeat="key in getKeys(json) track by $index">' + 
           '    <div ng-if="isPrimFn(json[key])"><span class=ajtd-key>{{ key + ": "}}</span><span class=ajtd-value>{{json[key]}}</span></div>' + 
           '    <div ng-if="!isPrimFn(json[key])">'+
           '      <div class=ajtd-array-key>{{ key }}</div>'+
           '      <div class=ajtd-object-value mn-json-tree-tweaked=json[key]></div>' + 
           '    </div>' + 
           '  </li>' + 
           '</ul>'      
       );
     }
     ]
);